////////////////////////////////////////////////////////////////////////////////
//
// Time-stamp: "10.08.11 00:34 AVR_LCD44780_fixpoint.cpp klaus?wachtler.de"
//
////////////////////////////////////////////////////////////////////////////////
//
// Testprogramm fr LCD auf Pollin-Addon-Board zum
// Pollin-AVR-Evaluationboard oder frei verdrahtet im 4-bit-Modus
// lt. www.mikrocontroller.net.
//
// Auf dem Addon-Board ist ein LCD 2x16 mit HDC44780.
// Anschlu (Jumperleiste 1:1 gesteckt):
// PortA0..7 -> DB0..7 an LCD
// PortB0 -> E an LCD
// PortB1 -> R/W an LCD
// PortB2 -> RS an LCD
//
// Folgende Kabel auf Addon-Board stecken:
// - PA1 (ATMega8: PC0)   -> LCD DB4
// - PA2 (ATMega8: PC1)   -> LCD DB5
// - PA3 (ATMega8: PC2)   -> LCD DB6
// - PA4 (ATMega8: PC3)   -> LCD DB7
// - PA5 (ATMega8: PC4)   -> LCD RS
// - PA6 (ATMega8: PC5)   -> LCD E
// - GND                  -> R/W
//
// Bei freier Verdrahtung mssen folgende Verbindungen existieren:
//           LCD                                            AVR
//       VSS  1-----GND
//       VDD  2-----+5V
//       VEE  3--Abgriff Poti 10k zw. GND und +5V
//        RS  4---------------------------------------------LCD-RS PC4
//       R/W  5-----GND
//         E  6---------------------------------------------LCD-E  PC5
//       DB0  7-NC
//       DB1  8-NC
//       DB2  9-NC
//       DB3 10-NC
//       DB4 11---------------------------------------------LCD-DB4 PC0
//       DB5 12---------------------------------------------LCD-DB5 PC1
//       DB6 13---------------------------------------------LCD-DB6 PC2
//       DB7 14---------------------------------------------LCD-DB7 PC3
//     LED-A 15--R 220Ohm--+5V
//     LED-K 15-----GND
//
//
//   ATMega8:
//
//                   +---------+
//     (/Reset)  PC6-|  1   28 |-PC5 (ADC5/SCL)
//        (RxD)  PD0-|  2   27 |-PC4 (ADC4/SDA)
//        (TxD)  PD1-|  3   26 |-PC3 (ADC3)
//       (INT0)  PD2-|  4   25 |-PC2 (ADC2)
//       (INT1)  PD3-|  5   24 |-PC1 (ADC1
//     (XCK/T0)  PD4-|  6   23 |-PC0 (ADC0)
//               VCC-|  7   22 |-GND
//               GND-|  8   21 |-AREF
// (XTAL1/TOSC1) PB6-|  9   20 |-AVCC
// (XTAL2/TOSC2) PB7-| 10   19 |-PB5 (SCK)
//          (T1) PD5-| 11   18 |-PB4 (MISO)
//        (AIN0) PD6-| 12   17 |-PB3 (MOSI/OC2)
//        (AIN1) PD7-| 13   16 |-PB2 (SS/OC1B)
//        (ICB1) PB0-| 14   15 |-PB1 (OC1A)
//                   +---------+
//
//
// Gesetzte Fuses:
//
// avrdude -c stk200 -p m8  -P /dev/parport0 -U hfuse:r:high.txt:b
// -> 0b11001001
//         + CKOPT=0
//
// avrdude -c stk200 -p m8  -P /dev/parport0 -U lfuse:r:low.txt:b
// -> 0b11111111 (CKSEL3..1 = 111)
//
// http://www.mikrocontroller.net/articles/AVR-GCC-Tutorial
//
////////////////////////////////////////////////////////////////////////////////
//
// Dieses Programm luft (laufzeit_sec) Sekunden.
// In dieser Zeit wird eine Schleife mit Berechnungen ausgefhrt und
// die Anzahl der Durchlufe gezhlt.
// Anschlieend wird die Anzahl/Sekunde auf dem LCD ausgegeben.
//
// Es kann jeweils einer der untenstehenden Datentypen entkommentiert
// werden.
//
// Solange man sich auf einen der Typen double bis uint... beschrnkt,
// kann der Quelltext ohne weitere nderung als C kompiliert werden;
// als C++ sind alle Datentypen mglich.
//
////////////////////////////////////////////////////////////////////////////////
//
// Ermittelte Laufzeiten und Codegre mit einem AtMega8 mit 16 MHz:
//
//                                  |     n/sec     |    .text    |
//                                  |  C    |  C++  |  C   |  C++ |
//  --------------------------------+-------+-------+------+------+
//  double                          |  3468 |  3468 | 4876 | 4876 |
//  --------------------------------+-------+-------+------+------+
//  uint8_t                         | 73487 | 73487 | 4312 | 4312 |
//  uint16_t                        | 51721 | 51721 | 4348 | 4348 |
//  uint32_t                        | 19614 | 19614 | 4500 | 4500 |
//  uint64_t                        |  1038 |  1038 | 6524 | 6524 |
//  --------------------------------+-------+-------+------+------+
//  fixpoint<uint8_t,uint8_t,1>     |       | 13161 |      | 4584 |
//  fixpoint<uint8_t,uint16_t,1>    |       | 31907 |      | 4552 |
//  fixpoint<uint16_t,uint16_t,1>   |       | 30300 |      | 4560 |
//  fixpoint<uint16_t,uint32_t,1>   |       | 12859 |      | 4726 |
//  fixpoint<uint32_t,uint32_t,1>   |       | 12629 |      | 4760 |
//  fixpoint<uint32_t,uint64_t,1>   |       |   626 |      | 6528 |
//  fixpoint<uint64_t,uint64_t,1>   |       |   597 |      | 7206 |
//  --------------------------------+-------+-------+------+------+
//
////////////////////////////////////////////////////////////////////////////////


//#include <stdlib.h>
//#include <stddef.h>
#include <stdio.h>
//#include <string.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdint.h>
#include <inttypes.h>
#include <util/delay.h>

#include "lcd-routines.h"

#include "fixpoint.h"

// Eine der folgenden Zeilen aktivieren, den Rest auskommentieren:

//typedef double                                      test_t;  typedef double   const_t; const char*t_text="double";
//typedef uint8_t                                     test_t;  typedef uint8_t  const_t; const char*t_text="uint8_t";
//typedef uint16_t                                    test_t;  typedef uint16_t const_t; const char*t_text="uint16_t";
//typedef uint32_t                                    test_t;  typedef uint32_t const_t; const char*t_text="uint32_t";
//typedef uint64_t                                    test_t;  typedef uint64_t const_t; const char*t_text="uint64_t";
//typedef Anyware::fixpoint< uint8_t, uint8_t, 1 >    test_t;  typedef uint8_t  const_t; const char*t_text="f(8,8)";
//typedef Anyware::fixpoint< uint8_t, uint16_t, 1 >   test_t;  typedef uint8_t  const_t; const char*t_text="f(8,16)";
//typedef Anyware::fixpoint< uint16_t, uint16_t, 1 >  test_t;  typedef uint16_t const_t; const char*t_text="f(16,16)";
typedef Anyware::fixpoint< uint16_t, uint32_t, 1 >  test_t;  typedef uint16_t const_t; const char*t_text="f(16,32)";
//typedef Anyware::fixpoint< uint32_t, uint32_t, 1 >  test_t;  typedef uint32_t const_t; const char*t_text="f(32,32)";
//typedef Anyware::fixpoint< uint32_t, uint64_t, 1 >  test_t;  typedef uint32_t const_t; const char*t_text="f(32,64)";
//typedef Anyware::fixpoint< uint64_t, uint64_t, 1 >  test_t;  typedef uint64_t const_t; const char*t_text="f(64,64)";


volatile uint8_t       bMessungLaeuft  =  0;
const int              laufzeit_sec    = 10;

#ifdef __cplusplus
const char *lang_text = "C++";
#else
const char *lang_text = "C";
#endif /* ifdef __cplusplus */



int main( int nargs, char **args )
{
  char          tmp_puffer[30];
  uint32_t      nGeschafft = 0;
  // Werte, mit denen etwas gerechnet werden soll:
  test_t       a = (const_t)3, b = (const_t)1, c = (const_t)1;

  lcd_init();

  // Timer 0 mit 1/1024 F_CPU starten
  // Ich takte mit 16 MHz, also wird der Zhler mit 16 kHz
  // hochgezhlt.
  // Nach je 256 Werten tritt ein Interrupt auf, also mit 16 kHz/256 =
  // 62 Hz.
  sei();
  TCCR0 |= (1<<CS00)|(1<<CS02);
  TIMSK = 1<<TOIE0;


  lcd_clear();
  lcd_set_cursor( 0, 1 );
  lcd_string( "Bitte warten..." );



  bMessungLaeuft = 1;
  while( bMessungLaeuft )
  {
    size_t       i;
    const size_t nLoop = 10;
    a = (const_t)3;
    b = (const_t)1;
    c = (const_t)1;

    for( i=0; i<nLoop; ++i )
    {
      if( (i%4)<2 )
      {
        b *= (const_t)2;
        a += b;
        c += a*b;
      }
      else
      {
        c -= a*b;
        a -= b;
        b /= (const_t)2;
      }
    }
    ++nGeschafft; // Durchlufe zhlen
  }

  lcd_clear();
  lcd_set_cursor( 0, 1 );
  sprintf( tmp_puffer, "%.0f %s %s", (double)c, t_text, lang_text );
  lcd_string( tmp_puffer );
  lcd_set_cursor( 0, 2 );
  sprintf( tmp_puffer, "%.1f / sec  ", nGeschafft/(double)laufzeit_sec );
  lcd_string( tmp_puffer );

  while( "warten aufs jngste Gericht" )
  {
  }

  return 0;
} /* main( int nargs, char **args ) */



ISR(TIMER0_OVF_vect)
{
  // Dieser Interrupt tritt mit 62 Hz auf.
  static uint16_t count = 0;

  if( ++count>=62*laufzeit_sec )
  {
    // nach 62*laufzeit_sec Durchlufen sind laufzeit_sec Sekunden um
    count = 0;

    // Messung anhalten
    bMessungLaeuft = 0;
  }

}
