////////////////////////////////////////////////////////////////////////////////
//
// Time-stamp: "10.08.11 00:48 generate_sin_table.cpp klaus?wachtler.de"
//
////////////////////////////////////////////////////////////////////////////////
//
//
//     Generate sin) for a specific instance of Anyware::fixpoint<> template
//     ---------------------------------------------------------------------
//
// This program generates C++ source code to specialize the sin()
// function template of Anyware::fixpoint for a specific set of
// parameters.
//
// The generated source code ist written to stdout.
//
////////////////////////////////////////////////////////////////////////////////
//
// (C) 2011
//     Klaus Wachtler
//     Breidingstr. 17
//     D-29614 Soltau
//     Phone: +49-171-4553039
//     Mail:  fixpoint<you guess the character missing here?>wachtler.de
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 3 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
//
////////////////////////////////////////////////////////////////////////////////
//
// Parameters
/////////////
//
// Call this program with 6 parameters where the first four of them
// give the desired Anyware::fixpoint instance:
// - <TBASE>          names the base type of the Anyware::fixpoint
//                    template, e.g. int16_t
// - <TTEMP>          names the type of intermediate calculation of the
//                    Anyware::fixpoint template, e.g. int32_t
// - <LFRACTIONAL>    the scale value (as binary exponent) of the
//                    Anyware::fixpoint template, e.g. 8
// - <ROUNDTONEAREST> true or false
// - <step>           gives the step width of the table in [rad],
//                    e.g. 0.1
//                    Small values result in large table, small values
//                    might lead to inaccurate results.
//                    The value should be adequate to the accuracy of
//                    the fixpoint values.
//                    E.g. LFRACTIONAL=2 makes a resolution of 0.25
//                    for the angle and the result as well. So a value
//                    of less than 0.25 has no advantage (and will be
//                    rounded up by this program).
//                    The value given at startup will be adjusted to
//                    match the internal resolution of the fixpoint
//                    type.
//  - <filename>      file to be written or - to write to stdout
//
////////////////////////////////////////////////////////////////////////////////
//
// History:
///////////
//
// 03.08.2011 kw     created
//
// 06.08.2011 kw     - more comments
//                   - using only a table [0..pi/2] instead [0..pi]
//                   - now DEBUG_FIXPOINT_CHECK_RANGE and
//                     DEBUG_FIXPOINT_FOLLOW_DOUBLE respected
//                   - in some cases no interpolation needed
//                   - interpolation fixed
//                   - the generated table is now commented
//
//  kw     
//
//  kw     
//
//  kw     
//
////////////////////////////////////////////////////////////////////////////////
//
// Build:
/////////
//
// g++ -Wall -Wextra generate_sin_table.cpp -o generate_sin_table
//
// Run:
///////
//
// ./generate_sin_table int16_t int32_t 8 true 0.1 -
//
////////////////////////////////////////////////////////////////////////////////
//
// TODO:
////////
//
// - guess and comment accuracy
//
//
////////////////////////////////////////////////////////////////////////////////
//
// [1] fixpoint.h
//     TODO link to http://www.wachtler.de
//
////////////////////////////////////////////////////////////////////////////////


#include <cstddef>
#include <cmath>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <ctime>
#include <sstream>
#include <cmath>

void usage( const char *prgName )
{
  std::cerr << prgName << std::endl;
  std::cerr << "" << std::endl;
  std::cerr << "This program generates a sin() function as C++ source code" << std::endl;
  std::cerr << "for an instance of Anyware::fixpoint<> template class." << std::endl;
  std::cerr << "" << std::endl;
  std::cerr << "Call it like this:" << std::endl;
  std::cerr << prgName
            << " <TBASE> <TTEMP> <LFRACTIONAL> <ROUNDTONEAREST> <stepwidth> <file>"
            << std::endl;
  std::cerr << "" << std::endl;
  std::cerr << "For more information on <TBASE> <TTEMP> <LFRACTIONAL> <ROUNDTONEAREST>"
            << std::endl;
  std::cerr << "see documentation of Anyware::fixpoint." << std::endl;
  std::cerr << "Unsigned types are obviously not useful." << std::endl;
  std::cerr << "" << std::endl;
  std::cerr << "<stepwidth> is the step width of the interpolation table. This value will be"
            << std::endl;
  std::cerr << "adjusted automatically to the nearest step fitting to <LFRACTIONAL>."
            << std::endl;
  std::cerr << "A stepwidth value of more than 0.25 makes no sense and will be rejected,"
            << std::endl;
  std::cerr << "as well as a LFRACTIONAL less than 2." << std::endl;
  std::cerr << "" << std::endl;
  std::cerr << "<file> is the output file or \"-\" for standard output." << std::endl;
  std::cerr << "" << std::endl;
  std::cerr << "See" << std::endl;
  std::cerr << " - http://www.wachtler.de/#fixpoint" << std::endl;
  std::cerr << " - http://www.mikrocontroller.net/topic/143537" << std::endl;
  std::cerr << " - http://www.mikrocontroller.net/topic/227823" << std::endl;
  std::cerr << "" << std::endl;
  std::cerr << "(C) 2011 Klaus Wachtler" << std::endl;

}

int main( int nargs, char **args )
{
  // check arguments

  if( nargs!=7 )
  {
    usage( args[0] );
    return 1;
  }

  // args
  std::string          sTbase( args[1] );
  std::string          sTtemp( args[2] );
  std::string          sLfractional( args[3] );
  std::string          sRoundtonearest( args[4] );
  std::string          sStep( args[5] );
  std::string          sFilename( args[6] );

  double               requestedStep;
  std::istringstream( sStep ) >> requestedStep;

  if( requestedStep>0.25 )
  {
    std::cerr << "The requested step " << requestedStep << " is too large." << std::endl;
    usage( args[0] );
    return 1;
  }

  int            lFractional;
  std::istringstream( sLfractional ) >> lFractional;

  if( lFractional<2 )
  {
    std::cerr << "The requested LFRACTIONAL " << lFractional << " is too small." << std::endl;
    usage( args[0] );
    return 1;
  }


  // fixpoint type the sin() will be specialized for
  const std::string    sType( std::string( "fixpoint< " )
                              + sTbase
                              + ", "
                              + sTtemp
                              + ", "
                              + sLfractional
                              + ", "
                              + sRoundtonearest
                              + " >"
                              );

  // actual time
  std::time_t       actTime;
  struct tm         structActtime;
  actTime = std::time( NULL );
  structActtime = *localtime( &actTime );

  std::ostream  &out = ( sFilename=="-"
                         ? std::cout
                         : *new std::ofstream( sFilename.c_str() )
                         );

  // choose real step width
  //
  // The width will be taken as (2^(-stepFractional)) to enable a fast
  // finding of the next sampling point in the table.
  int stepFractional;
  std::frexp( 0.5*requestedStep, &stepFractional );
  stepFractional = -stepFractional;

  // A width of less than 2^(-LFRACTIONAL) will be rounded up to that
  // value
  if( stepFractional > lFractional )
  {
    stepFractional = lFractional;
  }

  double realStep = std::pow( 2.0, -stepFractional );


  // write head of file
  out << "// " << sFilename << std::endl;
  out << "// Time-stamp: \""
      << std::setw( 2 ) << std::setfill( '0' ) << structActtime.tm_mday
      << "."
      << std::setw( 2 ) << std::setfill( '0' ) << structActtime.tm_mon+1 << "."
      << structActtime.tm_year+1900
      << " "
      << std::setw( 2 ) << std::setfill( '0' ) << structActtime.tm_hour << ":"
      << std::setw( 2 ) << std::setfill( '0' ) << structActtime.tm_min
      << " "
      << sFilename
      << " "
      << "klaus?wachtler.de\""
      << std::endl;
  out << "//" << std::endl;
  out << "// sin() interpolation from table for " << sType << std::endl;
  out << "// Requested step width is " << requestedStep << std::endl;
  out << "// Choosen step width is " << realStep << std::endl;
  out << "//" << std::endl;
  out << "// generated by " << args[0]
      << " "
      << std::setw( 2 ) << std::setfill( '0' ) << structActtime.tm_mday
      << "."
      << std::setw( 2 ) << std::setfill( '0' ) << structActtime.tm_mon+1 << "."
      << structActtime.tm_year+1900
      << " "
      << std::setw( 2 ) << std::setfill( '0' ) << structActtime.tm_hour << ":"
      << std::setw( 2 ) << std::setfill( '0' ) << structActtime.tm_min
      << std::endl;
  out << "//" << std::endl;
  out << std::endl;
  out << "#include <assert.h>" << std::endl;
  out << "#include <cmath>" << std::endl;
  out << std::endl;
  out << "namespace Anyware" << std::endl;
  out << "{" << std::endl;
  out << "  template<>" << std::endl;
  out << "  " << sType << std::endl;
  out << "  sin( const " << sType <<  " &angle_ )" << std::endl;
  out << "  {" << std::endl;
  out << "    " << sType <<  "   angle( angle_ );" << std::endl;
  out << std::endl;
  out << "#ifdef DEBUG_FIXPOINT_CHECK_RANGE" << std::endl;
  out << "    angle.checkRange( std::sin( double(angle) )," << std::endl;
  out << "                      \"sin() spec. implementation <"
      << sTbase << "," << sTtemp << "," << sLfractional << "," << sRoundtonearest << ">\","
      << std::endl;
  out << "                      __FILE__, __LINE__, __FUNCTION__, NULL" << std::endl;
  out << "                      );" << std::endl;
  out << "#endif /* ifdef DEBUG_FIXPOINT_CHECK_RANGE */" << std::endl;
  out << std::endl;
  out << "    " << sType << "    ret;" << std::endl;
  out << std::endl;
  out << "#ifdef DEBUG_FIXPOINT_FOLLOW_DOUBLE" << std::endl;
  out << "    ret.followExact = std::sin( angle.followExact );" << std::endl;
  out << "#endif /* ifdef DEBUG_FIXPOINT_FOLLOW_DOUBLE */" << std::endl;
  out << std::endl;
  out << "    // table with basevalue members of the sin values" << std::endl;
  out << "    // (from 0.0 to more than pi/2 in angle steps of " << realStep << ")" << std::endl;
  out << "    static " << sTbase << " table[] =" << std::endl;
  out << "      {" << std::endl;

  size_t    iValue = 0;
  double    angle =  0.0;
  do
  {
    double     sinValue = std::sin( angle );

    out << "        "
        << std::setfill( ' ' ) << std::setw( 10 ) << (int64_t)( sinValue*pow( 2.0, lFractional )+0.5 )
        << ", // ["
        << std::setfill( ' ' ) << std::setw( 3 ) << iValue
        << "]  sin("
        << std::fixed << std::setprecision( 8 ) << angle
        << ") = " << std::setprecision( 6 ) << sinValue
        << std::endl;
    iValue++;
    angle += realStep;
  }
  while( angle < M_PI/2.0+realStep );
  out.unsetf( std::ios_base::fixed );

  out << "      }; // table[]" << std::endl;
  out << std::endl;
  out << "    bool inverse = false;  // change of sign necessary?" << std::endl;
  out << std::endl;
  out << "    // move angle to [0..]" << std::endl;
  out << "    if( angle.getBasevalue()<0 )" << std::endl;
  out << "    {" << std::endl;
  out << "      angle = -angle;" << std::endl;
  out << "      inverse = !inverse;" << std::endl;
  out << "    }" << std::endl;
  out << std::endl;
  out << "    // move angle to [0..2*pi]" << std::endl;
  out << "    " << sType << "   period( 2.0*M_PI );" << std::endl;
  out << "    if( angle>period )" << std::endl;
  out << "    {" << std::endl;
  out << "      angle = (angle%period);" << std::endl;
  out << "    }" << std::endl;
  out << std::endl;
  out << "    // move angle to [0..pi]" << std::endl;
  out << "    period /= (" << sType << ")(" << sTbase << ")2;" << std::endl;  // now pi" << std::endl;
  out << "    if( angle>period )" << std::endl;
  out << "    {" << std::endl;
  out << "      angle -= period;" << std::endl;
  out << "      inverse = !inverse;" << std::endl;
  out << "    }" << std::endl;
  out << std::endl;
  out << "    // move angle to [0..pi/2]" << std::endl;
  out << "    if( angle.getBasevalue()*2>period.getBasevalue() )" << std::endl;
  out << "    {" << std::endl;
  out << "      angle = period - angle;" << std::endl;
  out << "    }" << std::endl;
  out << std::endl;
  out << "    // internal value of the angle" << std::endl;
  out << "    " << sTtemp << "   angleBasevalue = angle.getBasevalue();" << std::endl;
  out << std::endl;

  out << "    // angleBasevalue counts in steps of 2^(-"+sLfractional+")," << std::endl;
  out << "    // while the samples in the table above have a step width"  << std::endl;
  out << "    // of " << realStep << " or 2^(-" << stepFractional << ")." << std::endl;
  if( lFractional==stepFractional )
  {
    out << "    // Thus angleBasevalue is the samplepoint's index." << std::endl;
    out << "    assert( angleBasevalue<(" << sTtemp << ")(sizeof(table)/sizeof(table[0])) );" << std::endl;
    out << "    ret.setBasevalue( table[angleBasevalue] );" << std::endl;
  }
  else
  {
    out << "    // Thus angleBasevalue>>" << (lFractional-stepFractional)
        << " is the left samplepoint's index." << std::endl;
    out << "    " << sTtemp << "   leftIndex   = angleBasevalue>>"
        << (lFractional-stepFractional) << ";" << std::endl;
    out << "    assert( leftIndex<(" << sTtemp << ")(sizeof(table)/sizeof(table[0])) );" << std::endl;
    out << "    " << sTtemp << "   leftSample  = table[leftIndex];" << std::endl;
    out << "    if( !(angleBasevalue&0x"
        << std::hex << (1<<((lFractional-stepFractional)))-1 << ") )"
        << std::endl;
    out << "    {" << std::endl;
    out << "      // no interpolation needed because leftIndex hit exactly" << std::endl;
    out << "      ret.setBasevalue( leftSample );" << std::endl;
    out << "    }" << std::endl;
    out << "    else" << std::endl;
    out << "    {" << std::endl;
    out << "      assert( leftIndex+1<(" << sTtemp << ")(sizeof(table)/sizeof(table[0])) );" << std::endl;
    out << "      " << sTtemp << "   rightSample = table[leftIndex+1];" << std::endl;
    out << "      if( leftSample-1>=rightSample && leftSample+1<=rightSample )"
        << std::endl;
    out << "      {" << std::endl;
    out << "        // no interpolation needed because sin values are equal on both sides" << std::endl;
    out << "        ret.setBasevalue( leftSample );" << std::endl;
    out << "      }" << std::endl;
    out << "      else" << std::endl;
    out << "      {" << std::endl;
    out << "        // do some linear interpolation" << std::endl;
    out << "        ret.setBasevalue( leftSample + (((angleBasevalue&0x"
        << std::hex << (1<<((lFractional-stepFractional)))-1
        << ")*2+1)*(rightSample-leftSample)/0x" << (1<<(lFractional-stepFractional+1)) << std::dec << ") );"
        << std::endl;
    out << "      }" << std::endl;
    out << "    }" << std::endl;
  }
  out << std::endl;
  out << "    return ( inverse ? -ret : ret );" << std::endl;
  out << "  } // sin()" << std::endl;
  out << "} // namespace Anyware" << std::endl;
  out << std::endl;

  return 0;
}
